/*  liblipsevent - Library of LiPS event model
 * event.h - 
 *
 * Authors: ZHAO Liangjing <liangjing.zhao@orange-ftgroup.com> 
 * 
 * This software and associated documentation files (the "Software") 
 * are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 * All Rights Reserved. 
 *
 * A copyright license is hereby granted for redistribution and use of 
 * the Software in source and binary forms, with or without modification, 
 * provided that the following conditions are met: 
 * - Redistributions of source code must retain the above copyright notice, 
 * this copyright license and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright 
 * notice, this copyright license and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution. 
 * - Neither the name of LiPS  nor the names of its Members may be used 
 * to endorse or promote products derived from the Software without 
 * specific prior written permission. 
 *
 * A patent license for any Necessary Claims owned by Members of LiPS Forum 
 * to make, have made, use, import, offer to sell, lease and sell or otherwise 
 * distribute any implementation compliant with the any specification adopted 
 * by the LiPS Forumcan be obtained from the respective Members on reasonable 
 * and non-discriminatory terms and conditions and under reciprocity, as 
 * regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 * "AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 * AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 * ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 */

/* !
 *@file <event.h>
 *
 *@brief Public header file for the Event type.
 *@author ZHAO Liangjing
 *
 *****************************************************************************/

#ifndef LIPS_EVENT_H_
#define LIPS_EVENT_H_

#ifdef __cplusplus
extern "C"
{
#endif                          /* __cplusplus */

  /******************************************************************************
   * Includes
   *****************************************************************************/
#include "evt_type.h"

  /******************************************************************************
   * Type definition
   *****************************************************************************/

  /* ! @var typedef uint32_t err_code_t;
     @brief event error type

     Error code for event return to applications.
   */
  typedef uint32_t err_code_t;

  /* ! @var typedef uint32_t evt_err_t;
     @brief event error type

     No additional information.
   */
  typedef uint32_t evt_err_t;

  /* ! @var typedef uint16_t evt_group_t;
     @brief event group type

     No additional information.
   */
  typedef uint16_t evt_group_t;

  /* ! @var typedef uint16_t evt_type_t
     @brief event type.

     No additional information.
   */
  typedef uint16_t evt_type_t;

  /* ! @var typedef struct _req_state_t req_state_t
     @brief asynchronous event state type.

     The state can only be applied to asynchronous events.
   */
  typedef _req_state_t req_state_t;

  /* ! @var typedef struct _evt_t evt_t
     @brief Event structure

     This structure is the standard event struct.
   */
  typedef struct _evt_t evt_t;

  /* ! @typedef void (*cb_t) (sid_t, evt_t*, void*);
     @brief Callback

     No additional information.
   */
  typedef void (*cb_t) (sid_t, evt_t *, void *userdata);

  /* ! @typedef void (*evt_data_free_t) (evt_t*);
     @brief event data free function

     No additional information.
   */
  typedef evt_err_t (*evt_data_free_t) (void *);

  /* ! @struct evt_handler_t
     @brief Event Handler data structure

     This structure contains a pointer to a callback as well as the user data to be passed
   */
  typedef struct
  {
    cb_t cb;                  /**< Callback */
    void *user_data;          /**< Private user data */
  }
  evt_handler_t;

  /* ! @struct evt_async_t
     @brief Event Asynchronous data structure

     This structure is used in tel_evt_t struct.
   */
  typedef struct
  {
    err_code_t err_code;         /**< Error Code */
    rid_t rid;                   /**< Request Identifier */
    req_state_t state;           /**< Async request state (notify/cancelled/complete) */
  }
  evt_async_t;

  /* ! @struct evt_t
     @brief Event structure

     This structure is the standard event struct.
   */
  struct _evt_t
  {
    evt_type_t evt_type;                /**< Event ID */
    sid_t sid;                          /**< Session ID */
    evt_handler_t *evt_handler;         /**< Call back & user data pointers */
    void *evt_data;                     /**< Event Data */
    evt_async_t *async_data;            /**< Asynchronous related information */
    evt_data_free_t evt_data_free;      /**< Pointer to the event data free function */
  };

  //void evt_t_free(evt_t*evt);

  /* ! @var typedef struct _evt_src_t evt_src_t
     @brief an opaque structure to encapsulate an event source.

     An application can contains a couple of enablers, so that
     it will have multiple event sources. The events come from
     different event sources (means different enablers) could
     have the same session id. But it does not mean the events
     come from the same session.
   */
  typedef struct _evt_src_t evt_src_t;

  /**
   *Register an unsolicited event.
   *
   *For catching the unsolicited event, application should register
   *the callback function and the expected event type with this function.
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param type expected event type.
   *@param handler the handler pointer corresponding to the event.
   *
   *@return return an #evt_err_t value.
   *                      - #EVT_ERROR_NONE: Successful, no error;
   *                      - #EVT_ERROR_DUPLICATED_REG: failed, the event is already registered.
   */
  evt_err_t evt_register (evt_src_t *evt_src,
                          sid_t sid, evt_type_t type, evt_handler_t *handler);

  /**
   *Unregister an unsolicited event.
   *
   *If application no more need to catch an unsolicited event, it should unregister
   *the callback function and the event type with this function.
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param type expected event type.
   *
   *@return return an #evt_err_t value.
   */
  evt_err_t evt_unregister (evt_src_t *evt_src, sid_t sid, evt_type_t type);

  /**
   *@fn evt_err_t (*get_evt_fd_t)(evt_src_t *evt_src, sid_t sid, int*fd)
   *@brief Get a regular file descriptor for querying event .
   *
   *This function provide a regular file descriptor that enable the client
   *to check pending events using a select() (or a poll()) system call.
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param[out] fd return the regular file descriptor.
   *
   *@return return an #evt_err_t value.
   */
  typedef evt_err_t (*get_evt_fd_t) (evt_src_t *evt_src, sid_t sid, int *fd);

  /**
   *Proccess & deliver events .
   *
   *This function delivers an event to the client by calling the callback function
   *specified through the corresponding #evt_register request. The callback
   *function is called in the context of the evt_process request.
   *
   *This function supports three forms: blocking, non-blocking and timed-out. The
   *blocking form may block for an undetermined amount of time waiting for an event.
   *The non-blocking form always returns immediately either delivering a pending
   *event or with an error code when there is no pending events to deliver. The timed-out
   *form may block waiting for an event but will return with an error if there is no
   *events to deliver within a specified time interval.
   *
   *@see evt_get
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param timeout an integer indicates the timeout duration:
   *                         - <b>0</b>: non-blocking form;
   *                         - <b>-1</b>: blocking form;
   *                         - <b>a positive value</b>: timed-out form, indicates the block
   *                                                                   time interval in millisecond.
   *
   *@return return an #evt_err_t value.
   */
  evt_err_t evt_process (evt_src_t *evt_src, sid_t sid, int timeout);

  /**
   *@fn evt_err_t (*evt_get_t)(evt_src_t *evt_src, sid_t sid, evt_t**evt, int timeout)
   *@brief Comment part for Get an events.
   *
   *This function delivers a pending event to the client as an event_descriptor.
   *Once delivered the client may access any information contained in the descriptor.
   *
   *This function supports three forms: blocking, non-blocking and timed-out. The
   *blocking form may block for an undetermined amount of time waiting for an event.
   *The non-blocking form always returns immediately either delivering a pending
   *event or with an error code when there is no pending events to deliver. The timed-out
   *form may block waiting for an event but will return with an error if there is no
   *events to deliver within a specified time interval.
   *
   *This function should be implemented by users.
   *
   *@see evt_process
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param[out] evt return a pointer to a pending event.
   *                It should be released in #evt_release_t function.
   *@param timeout an integer indicates the timeout duration:
   *                         - <b>0</b>: non-blocking form;
   *                         - <b>-1</b>: blocking form;
   *                         - <b>a positive value</b>: timed-out form, indicates the block
   *                                                                   time interval in millisecond.
   *
   *@return return an #evt_err_t value.
   */
  typedef evt_err_t (*evt_get_t) (evt_src_t *evt_src, sid_t sid, evt_t **evt,
                                  int timeout);

  /**
   *@brief default #evt_get_t function behavior.
   *
   *a the #evt_get_t function can invoke this function
   *in the very beginning.
   *
   *This function will create an empty event structure, and
   *set the sid.
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param evt_type the type of the event.
   *@param evt_data_free the event data free function.
   *@param[out] evt return a pointer to a pending event.
   *                It should be released in #evt_release_t function.
   *
   *@see #evt_get_t
   */
  evt_err_t _evt_get (evt_src_t *evt_src, sid_t sid, evt_type_t evt_type,
                      evt_data_free_t evt_data_free, evt_t **evt);

  /**
   *Call the event callback function.
   *
   *This function delivers event simply by invoking the callback function contained
   *in the evt parameter.
   *
   *@param evt_src the event source.
   *@param evt the event descriptor representing the despatching event.
   *
   *@return return an #evt_err_t value.
   */
  evt_err_t evt_call (evt_src_t *evt_src, evt_t *evt);

  /**
   *@fn evt_err_t evt_release(evt_src_t *evt_src, sid_t sid, evt_t*evt)
   *@brief Release the resource related to an event.
   *
   *This function releases all resources the implementation (i.e. server) has
   *allocated for the event including the event descriptor it-self.
   *
   *This function should be implemented by users.
   *
   *@param evt_src the event source.
   *@param evt the event descriptor representing the releasing event.
   *           It should be free in this function.
   *
   *@return return an #evt_err_t value.
   */
  evt_err_t evt_release (evt_src_t *evt_src, evt_t *evt);

  /**
   *Register an asynchronous event.
   *
   *For catching the asynchronous event, application should register
   *the callback function and the expected event type with this function.
   *
   *@param evt_src the event source.
   *@param sid session id.
   *@param rid request id.
   *@param handler the handler pointer corresponding to the event.
   *
   *@return return an #evt_err_t value.
   *                      - #EVT_ERROR_NONE: Successful, no error;
   *                      - #EVT_ERROR_DUPLICATED_REG: failed, the event was already registered.
   */
  evt_err_t async_evt_register (evt_src_t *evt_src,
                                sid_t sid, rid_t rid, evt_handler_t *handler);

  /**
   *@brief search an unsolicited event handler related to specified session & event type.
   *
   *@param evt_src the event source.
   *@param sid the session id.
   *@param etype the event type.
   *@param[out] handler a pointer to return the handler.
   *
   *@return return an #evt_err_t value.
   *                      - #EVT_ERROR_NONE: Successful, no error;
   *                      - #EVT_ERROR_NO_SUCH_EVT: failed, the event was not registered.
   */
  evt_err_t evt_get_handler (evt_src_t *evt_src,
                             sid_t sid,
                             evt_type_t etype, evt_handler_t **handler);

  /**
   *@brief search an asynchronous event handler related to specified session & rid.
   *
   *@param evt_src the event source.
   *@param sid the session id.
   *@param rid the request id.
   *@param[out] handler a pointer to return the handler.
   *
   *@return return an #evt_err_t value.
   *                      - #EVT_ERROR_NONE: Successful, no error;
   *                      - #EVT_ERROR_NO_SUCH_EVT: failed, no handler for this request.
   */
  evt_err_t async_evt_get_handler (evt_src_t *evt_src,
                                   sid_t sid,
                                   rid_t rid, evt_handler_t **handler);

  /**
   *@brief initialize resource for using LiPS event library
   *
   *@param evt_get #evt_get_t function.
   *@param evt_release #evt_release_t function.
   *@param get_evt_fd #get_evt_fd_t function.
   *@param[out] pEvt_src a pointer to return an evt_src_t structure.
   *
   *@return return an #evt_err_t value.
   */
  evt_err_t evt_init (evt_get_t evt_get,
                      get_evt_fd_t get_evt_fd, evt_src_t **pEvt_src);

  /**
   *@brief free the resource used by an event source.
   *
   *@param evt_src the event source.
   *
   *@return return an #evt_err_t value.
   */
  evt_err_t evt_finalize (evt_src_t *evt_src);



#ifdef __cplusplus
}
#endif                          /* __cplusplus */

#endif                          // LIPS_EVENT_H_

