/*  liblipsevent - Library of LiPS event model
 * event.c - 
 *
 * Authors: ZHAO Liangjing <liangjing.zhao@orange-ftgroup.com> 
 * 
 * This software and associated documentation files (the "Software") 
 * are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 * All Rights Reserved. 
 *
 * A copyright license is hereby granted for redistribution and use of 
 * the Software in source and binary forms, with or without modification, 
 * provided that the following conditions are met: 
 * - Redistributions of source code must retain the above copyright notice, 
 * this copyright license and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright 
 * notice, this copyright license and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution. 
 * - Neither the name of LiPS  nor the names of its Members may be used 
 * to endorse or promote products derived from the Software without 
 * specific prior written permission. 
 *
 * A patent license for any Necessary Claims owned by Members of LiPS Forum 
 * to make, have made, use, import, offer to sell, lease and sell or otherwise 
 * distribute any implementation compliant with the any specification adopted 
 * by the LiPS Forumcan be obtained from the respective Members on reasonable 
 * and non-discriminatory terms and conditions and under reciprocity, as 
 * regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 * "AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 * AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 * ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 */

/* !
 *@file <event.c>
 *
 *@brief body file for Event type.
 *@author ZHAO Liangjing
 *
 *****************************************************************************/

#include "event.h"
#include <glib.h>

//key for unsolicited event register hash table.
typedef struct
{
  evt_type_t type;
  sid_t sid;
}
evt_key_t;

//key for asynchronous event register hash table.
typedef struct
{
  rid_t rid;
  sid_t sid;
}
req_key_t;

gboolean
evt_key_equal (gconstpointer v, gconstpointer v2)
{
  return (((evt_key_t *) v)->type == ((evt_key_t *) v2)->type) ?
    (((evt_key_t *) v)->sid == ((evt_key_t *) v2)->sid) : FALSE;
};

gboolean
req_key_equal (gconstpointer v, gconstpointer v2)
{
  return (((req_key_t *) v)->rid == ((req_key_t *) v2)->rid) ?
    (((req_key_t *) v)->sid == ((req_key_t *) v2)->sid) : FALSE;
};

guint
evt_key_hash (gconstpointer v)
{
  guint32 h = (0x0000ffff & ((evt_key_t *) v)->sid) << 16;

  return h | (0x0000ffff & ((evt_key_t *) v)->type);
};

guint
req_key_hash (gconstpointer v)
{
  guint32 h = (0x0000ffff & ((req_key_t *) v)->sid) << 16;

  return h | (0x0000ffff & ((req_key_t *) v)->rid);
};

struct _evt_src_t
{
  GHashTable *evt_hs;
  GMutex *evt_mutex;
  GHashTable *req_hs;
  GMutex *req_mutex;
  get_evt_fd_t get_evt_fd;
  evt_get_t evt_get;
};

/**
 *@brief default get_evt_fd_t functions
 *
 *This function will always return #EVT_ERROR_NOSUPPORT error.
 *@see #get_evt_fd_t
 */
evt_err_t
_get_evt_fd (evt_src_t *evt_src, sid_t sid, int *fd)
{
  return EVT_ERROR_NOSUPPORT;
};

evt_err_t
evt_init (evt_get_t evt_get, get_evt_fd_t get_evt_fd, evt_src_t **pEvt_src)
{
  g_assert (evt_get);

  *pEvt_src = (evt_src_t *) g_new0 (evt_src_t, 1);
  if (!*pEvt_src)
    return EVT_ERROR_MEM;
  if (!g_thread_supported ())
    g_thread_init (NULL);

  (*pEvt_src)->evt_hs = g_hash_table_new_full (evt_key_hash,
                                               evt_key_equal, g_free, NULL);
  if (!(*pEvt_src)->evt_hs)
    return EVT_ERROR_MEM;
  (*pEvt_src)->evt_mutex = g_mutex_new ();
  if (!(*pEvt_src)->evt_mutex)
    return EVT_ERROR_MEM;

  (*pEvt_src)->req_hs = g_hash_table_new_full (req_key_hash,
                                               req_key_equal, g_free, NULL);
  if (!(*pEvt_src)->req_hs)
    return EVT_ERROR_MEM;
  (*pEvt_src)->req_mutex = g_mutex_new ();
  if (!(*pEvt_src)->req_mutex)
    return EVT_ERROR_MEM;

  (*pEvt_src)->evt_get = evt_get;
  if (get_evt_fd)
    (*pEvt_src)->get_evt_fd = get_evt_fd;
  else
    (*pEvt_src)->get_evt_fd = _get_evt_fd;      //use defaulte get_evt_fd_t function.

  return EVT_ERROR_NONE;
}

evt_err_t
evt_finalize (evt_src_t *evt_src)
{
  g_assert (evt_src);
  if (evt_src->evt_hs)
    g_hash_table_destroy (evt_src->evt_hs);
  if (evt_src->evt_mutex)
    g_mutex_free (evt_src->evt_mutex);
  if (evt_src->req_hs)
    g_hash_table_destroy (evt_src->req_hs);
  if (evt_src->req_mutex)
    g_mutex_free (evt_src->req_mutex);
  g_free (evt_src);
  return EVT_ERROR_NONE;
}

bool
evt_is_registered (evt_src_t *evt_src, evt_key_t *key)
{
  return (g_hash_table_lookup (evt_src->evt_hs, key) != NULL);
};

bool
req_is_registered (evt_src_t *evt_src, req_key_t *key)
{
  return (g_hash_table_lookup (evt_src->req_hs, key) != NULL);
};

evt_err_t
evt_register (evt_src_t *evt_src,
              sid_t sid, evt_type_t type, evt_handler_t *handler)
{
  g_assert (evt_src);
  evt_key_t *key = g_new0 (evt_key_t, 1);

  key->type = type;
  key->sid = sid;

  g_mutex_lock (evt_src->evt_mutex);
  if (!evt_is_registered (evt_src, key))
  {
    g_hash_table_insert (evt_src->evt_hs, key, handler);
    g_mutex_unlock (evt_src->evt_mutex);
    return EVT_ERROR_NONE;
  }
  g_mutex_unlock (evt_src->evt_mutex);
  g_free (key);
  return EVT_ERROR_DUPLICATED_REG;
};

evt_err_t
async_evt_register (evt_src_t *evt_src,
                    sid_t sid, rid_t rid, evt_handler_t *handler)
{
  g_assert (evt_src);
  req_key_t *key = g_new0 (req_key_t, 1);

  key->rid = rid;
  key->sid = sid;

  g_mutex_lock (evt_src->req_mutex);
  if (!req_is_registered (evt_src, key))
  {
    g_hash_table_insert (evt_src->req_hs, key, handler);
    g_mutex_unlock (evt_src->req_mutex);
    return EVT_ERROR_NONE;
  }
  g_mutex_unlock (evt_src->req_mutex);
  g_free (key);
  return EVT_ERROR_DUPLICATED_REG;
}

evt_err_t
evt_unregister (evt_src_t *evt_src, sid_t sid, evt_type_t type)
{
  evt_key_t key;

  key.type = type;
  key.sid = sid;

  g_mutex_lock (evt_src->evt_mutex);
  if (g_hash_table_remove (evt_src->evt_hs, &key))
  {
    g_mutex_unlock (evt_src->evt_mutex);
    return EVT_ERROR_NONE;
  }
  else
  {
    g_mutex_unlock (evt_src->evt_mutex);
    return EVT_ERROR_NULL;      // no such registered event.
  }
};

evt_err_t
pop_async_evt (evt_src_t *evt_src, sid_t sid, rid_t rid)
{
  req_key_t key;

  key.rid = rid;
  key.sid = sid;

  g_mutex_lock (evt_src->req_mutex);
  if (g_hash_table_remove (evt_src->req_hs, &key))
  {
    g_mutex_unlock (evt_src->req_mutex);
    return EVT_ERROR_NONE;
  }
  else
  {
    g_mutex_unlock (evt_src->req_mutex);
    return EVT_ERROR_NULL;      // no such registered event.
  }
};

evt_err_t
evt_process (evt_src_t *evt_src, sid_t sid, int timeout)
{
  g_assert (evt_src);
  evt_err_t rt = EVT_ERROR_NONE;
  evt_t *evt = NULL;

//   g_print("in %s:%s %d\n", __FILE__, __FUNCTION__, __LINE__);
  rt = evt_src->evt_get (evt_src, sid, &evt, timeout);
//   printf("-----------------\n");
//   g_print("in %s:%s %d\n", __FILE__, __FUNCTION__, __LINE__);
  if (rt != EVT_ERROR_NONE)
  {
//     g_print("evt_get failed! before evt_release\n");
    evt_release (evt_src, evt);
//     g_print("evt_get failed! after evt_release\n");
    return rt;
  }

//   g_print("before evt_call\n");
  rt = evt_call (evt_src, evt);
  //if(rt != EVT_ERROR_NONE) return rt;

  return evt_release (evt_src, evt);
};

evt_err_t
_evt_get (evt_src_t *evt_src, sid_t sid, evt_type_t evt_type,
          evt_data_free_t evt_data_free, evt_t **evt)
{
  *evt = g_new0 (evt_t, 1);
  if (!*evt)
    return EVT_ERROR_MEM;
  (*evt)->sid = sid;
  (*evt)->evt_type = evt_type;
  (*evt)->evt_data_free = evt_data_free;
  return EVT_ERROR_NONE;
};

evt_err_t
evt_call (evt_src_t *evt_src, evt_t *evt)
{
//   g_print("in evt_call\n");
  g_assert (evt);
  g_assert (evt->evt_handler);
  if (evt->evt_handler->cb)
  {
    (evt->evt_handler->cb) (evt->sid, evt, evt->evt_handler->user_data);
  }
  if (evt->async_data)
  {                             // asynchronous event
    pop_async_evt (evt_src, evt->sid, evt->async_data->rid);
  }
  return EVT_ERROR_NONE;
};

evt_err_t
evt_release (evt_src_t *evt_src, evt_t *evt)
{
  if (!evt)
    return EVT_ERROR_NULL;
  if (evt->async_data)
    g_free (evt->async_data);
  if (evt->evt_data && evt->evt_data_free)
  {                             //ensure evt_data & evt_data_free are not NULL
    (evt->evt_data_free) (evt->evt_data);
  }
  g_free (evt);
  return EVT_ERROR_NONE;
};

evt_err_t
evt_get_handler (evt_src_t *evt_src, sid_t sid, evt_type_t etype,
                 evt_handler_t **handler)
{
  evt_key_t *key = g_new0 (evt_key_t, 1);

  key->sid = sid;
  key->type = etype;
  *handler = g_hash_table_lookup (evt_src->evt_hs, key);
  if (*handler)
    return EVT_ERROR_NONE;
  return EVT_ERROR_NO_SUCH_EVT;
};

evt_err_t
async_evt_get_handler (evt_src_t *evt_src, sid_t sid, rid_t rid,
                       evt_handler_t **handler)
{
  req_key_t *key = g_new0 (req_key_t, 1);

  key->sid = sid;
  key->rid = rid;
  *handler = g_hash_table_lookup (evt_src->req_hs, key);
  if (*handler)
    return EVT_ERROR_NONE;
  return EVT_ERROR_NO_SUCH_EVT;
};

